// MonitorSession.java                                           -*- Java -*-
//    A class to encapsulate a session with a monitor
//
// COPYRIGHT (C) 1998, Bradley M. Kuhn
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//
// Written   :   Bradley M. Kuhn         University of Cincinnati
//   By          
//
// Written   :   John Franco
//   For         Special Topics: Java Programming
//               15-625-595-001, Fall 1998
// RCS       :
//
// $Source: /home/server/Project/RCS/MonitorSession.java $
// $Revision: 0.12 $
// $Date: 1998/12/15 05:56:01 $
//
// $Log: MonitorSession.java $
// Revision 0.12  1998/12/15 05:56:01  bkuhn
//   -- put files under the GPL
//
// Revision 0.11  1998/11/30 07:56:06  bkuhn
//    -- added methods to support encryption
//    -- made println and readLine so that I can hide the encryption
//       inside the MonitorSEssion
//
// Revision 0.10  1998/11/29 12:01:55  bkuhn
//   -- rewrote aliveCheck() to make it a bit more readable,
//      and fix the problem with checkConnectingHost() not sending an
//      error message properly.  I subsequently commented out the whole
//      check for hosts matching because it was unfair to those who
//      didn't have accounts on machines that held servers they sought
//      to impersonate
//
// Revision 0.9  1998/11/25 06:49:33  bkuhn
//   -- added the ability to get access to the current command from
//      other classes; I needed this for WarRunner
//   -- added the sendDirective() method, since I had send.* methods
//      for all the other things that got sent.  I am not using
//      the sendDirective() method everywhere, yet, though.
//
// Revision 0.8  1998/11/17 21:19:13  bkuhn
//   # moved constasnts to GameParameters
//
// Revision 0.7  1998/11/16 08:07:15  bkuhn
//   # increased version number to 0.9
//
// Revision 0.6  1998/11/15 17:45:24  bkuhn
//    # updated version number
//   -- fixed bug where both COMMAND_ERROR: and RESULT: was sent in Message Group
//    responses after some commands has been fixed.
//
// Revision 0.5  1998/11/13 08:37:53  bkuhn
//   -- added a RandomRecurringEvent for CheckForLiving
//
// Revision 0.4  1998/11/09 07:18:55  bkuhn
//   -- reorganized things so that the guts of command handling
//      is actually methods in here.  This allows the sub
//      classes to very easily require() things and process
//      commands.
//
// Revision 0.2  1998/11/03 06:15:53  bkuhn
//   # initial version
//

import java.io.*;
import java.net.*;

/*****************************************************************************/
class MonitorSession
{
      static final String rcsid = "$Revision: 0.12 $";

      BufferedReader  input;
      PrintStream     output;
      // Check it out!  PrintStream is deprecated, and one is supposed
      // to replace it with PrintWriter----PrintWriter works great 
      // in kaffe, but Sun's JVM can't do anything with it.
      // What do you think of that?
      
      Player          player, auxiliary;
      PlayerDB        playerDB;

      Command         curCommand;
      
      DHEngine        dhEngine;
      Cipher          cipher;
      
      /*********************************************************************/
      MonitorSession(PlayerDB players)
      {
         playerDB = players;

         dhEngine = null;
         cipher = null;
      }
      /*********************************************************************/
      boolean setDHExchangeKey(DHKeyObject dhKey, String key)
      {
         dhEngine = new DHEngine(playerDB.getDHKey());
         return dhEngine.setExchangeKey(key);
      }
      /*********************************************************************/
      String getDHExchangeKey()
      {
         return dhEngine.getExchangeKey();
      }
      /*********************************************************************/
      void enableCipherMode()
      {
         cipher = new Cipher(dhEngine.getSharedKey());
      }
      /*********************************************************************/
      void println(String s)
      {
         if (cipher == null)
            output.println(s);
         else
         {
            s = cipher.Encrypt(s);
            output.println(s);
         }
      }
      /*********************************************************************/
      String readLine()
         throws IOException
      {
          String s = input.readLine();

          if (cipher == null)
              return s;
          else
             return (s == null) ? s : cipher.Decrypt(s);
      }
      /*********************************************************************/
      private PrintStream getOutput()
      {
         return output;
      }
      /*********************************************************************/
      void greet()
      {
         sendComment("Game Monitor Version " +
                     GameParameters.MONITOR_VERSION);
      }
      /*********************************************************************/
      boolean checkConnectingHost(String hostname)
      {
         if (player == null || player.getHostName() == null)
            return false;
         else
            return player.getHostName().equalsIgnoreCase(hostname);
      }         
      /*********************************************************************/
      void sendComplete()
      {
         println(Directive.WAITING_DIRECTIVE);
      }         
      /*********************************************************************/
      public boolean requireVerifyAndExecute(String commandRequried)
         throws IOException, MonitorSessionException
      {
         int count = 0;
         boolean notYetVerified = true;
         
         while (notYetVerified)
         {
            if (++count > GameParameters.MAX_COMMAND_RETRY_ON_REQUIRE)
            {
               throw new MonitorSessionException(
                  "Excessive command failures...Disconnecting...");
            }
            
            require(commandRequried);
                     
            if ( curCommand != null && curCommand.verify(this) ) 
            {
               curCommand.execute(this);
               notYetVerified = false;
            }
            curCommand = null;
         }            
         return true;
      }
      /*********************************************************************/
      void require(String commandRequired)
         throws IOException, MonitorSessionException
      {
         completeOutputAndGetCommand(commandRequired,
                                 GameParameters.MAX_COMMAND_RETRY_ON_REQUIRE);
      }                
      /*********************************************************************/
      void completeOutputAndGetCommand(int maxRetry)
         throws IOException, MonitorSessionException
      {
         int count = 0;

         while (curCommand == null)
         {
            if (++count > maxRetry)
            {
               throw new MonitorSessionException(
                  "Excessive command failures...Disconnecting...");
            }
            sendComplete();

            try
            { 
               curCommand = Command.getInstance(readLine());
            }
            catch (CommandException ce)
            {
               ce.report(this);
               curCommand = null;
            }
         }
      }
      /*********************************************************************/
      void completeOutputAndGetCommand(String commandRequired, int maxRetry)
         throws IOException, MonitorSessionException
      {
         int count = 0;

         while (curCommand == null)
         {
            if (++count > maxRetry)
               throw new MonitorSessionException(
                  "Excessive command failures...Disconnecting...");

            println(Directive.REQUIRE_DIRECTIVE + commandRequired);
            sendComplete();

            try
            { 
               curCommand = Command.getInstance(readLine(), commandRequired);
            }
            catch (CommandException ce)
            {
               ce.report(this);
               curCommand = null;
            }
         }
      }
      /*********************************************************************/
      Player requestIdent()
         throws IOException, MonitorSessionException
      {
         Player p = null;
         require(IdentCommand.COMMAND_STRING);
         if (curCommand != null && curCommand.verify(this))
         {
            curCommand.execute(this);
            p = playerDB.lookupOrCreate((IdentCommand) curCommand);
         }
         curCommand = null;

         if (p == null)
            sendError(" Unable to find or create player with given identity");
         return p;
      }
      /*********************************************************************/
      public PlayerDB getPlayerDB() 
      {
         return playerDB;
      }
      /*********************************************************************/
      public Player getPlayer() 
      {
         return player;
      }
      /*********************************************************************/
      public Player getAuxillaryPlayer() 
      {
         return auxiliary;
      }
      /*******************************************************************/
      public boolean aliveCheck(InetAddress address, int port, boolean
                                reportProblemsAsCommandError)
      {
         AliveConnection aliveConnection = null;
         boolean returnValue = true;

         if (player == null || playerDB == null)
            return false;

         try
         {
            aliveConnection = 
               new AliveConnection(address, port, playerDB, player);
         }            
         catch (MonitorSessionException mse)
         {
            sendErrorOrComment(reportProblemsAsCommandError,
                               mse.getMessage());
            returnValue =  false;
         }
         if (returnValue)
         {
            aliveConnection.start();
         
            try { aliveConnection.join(); }
            catch(InterruptedException ie) { }
         
            if (! aliveConnection.completedNormally())
               sendErrorOrComment(reportProblemsAsCommandError,
                                  "unable to verify server alive at "
                                  + address.getHostName() + " on port "
                                  + port);
            returnValue = aliveConnection.completedNormally();

// Put this code back in if you want to force that all active client
//  connects come from the same machine that the server comes from
//             if (! checkConnectingHost(address.getHostName()))
//             {
//                sendErrorOrComment(reportProblemsAsCommandError,
//                                   "Connection mismatch - " +
//                                   " your active connection is from a different"
//                                   + " machine than your server lives on.");

//                returnValue = false;
//             }
         }
         return returnValue;
      }
      /*******************************************************************/
      /* initiate() returns TRUE iff. we are able to initate a session
      ** with the player.  player will be set to a valid player if
      ** the session was initiated.
      */
      boolean initiate()
         throws MonitorSessionException
      {
         try
         {
            player = requestIdent();
         }
         catch (IOException ioe)
         {
            throw new MonitorSessionCreationException(
               "unable to identify player");
         }
         return (player != null);
      }
      /*********************************************************************/
      boolean getNewCommand()
         throws MonitorSessionException
      {
         if (curCommand == null) 
         {
            try
            {
               completeOutputAndGetCommand(GameParameters.MAX_COMMAND_RETRY);
            }
            catch (IOException ie)
            {
                curCommand = null;
                endSession();
                throw new MonitorSessionException("Session complete");
            }
         }
         return (curCommand != null);
      }            
      /*********************************************************************/
      Command getCurCommand()
      {
         return curCommand;
      }
      /*********************************************************************/
      void clearCurCommand()
      {
         curCommand = null;
      }
      /*********************************************************************/
      void handleCurrentCommand()
      {
         if (curCommand == null)
            return;

         if (curCommand.verify(this))
            curCommand.execute(this);

         curCommand = null;
      }
      /*********************************************************************/
      public void endSession()
      {
         try { input.close(); output.close(); } catch (IOException ie) { }
      }
      /*********************************************************************/
      public void sendResult(String result)
      {
         println(Directive.RESULT_DIRECTIVE + result);
      }
      /*********************************************************************/
      public void sendDirective(String directive, String data)
      {
         println(Directive.WAR_DECLARATION_DIRECTIVE + data);
      }
      /*********************************************************************/
      void sendComment(String s)
      {
         println(Directive.COMMENT_DIRECTIVE + s);
      }         
      /*********************************************************************/
      public void sendError(String error)
      {
         CommandException.report(this, error);
      }
      /*********************************************************************/
      public void sendErrorOrComment(boolean reportProblemsAsCommandError, 
                                     String message)
      {
         if (reportProblemsAsCommandError)
            sendError(message);
         else
            sendComment(message);
      }
}

/*
  Local Variables:
  tab-width: 4
  indent-tabs-mode: nil
  eval: (c-set-style "ellemtel")
  End:
*/
